from __future__ import annotations

import asyncio
import json

from textual.app import App
from ypywidgets import Widget as _Widget, reactive
from ypywidgets.comm import CommWidget

from ._driver import SIZE


class WidgetModel(_Widget):
    _data_from_app = reactive("")
    _data_to_app = reactive("")
    _cols = reactive(0)
    _rows = reactive(0)
    _ready = reactive(False)

    def __init__(
        self,
        ydoc=None,
    ) -> None:
        super().__init__(ydoc)


class Widget(CommWidget, WidgetModel):

    def __init__(self, app: App, cols: int = 0, rows: int = 0) -> None:
        CommWidget.__init__(self)
        WidgetModel.__init__(self)
        self._app = app
        SIZE[0] = self._cols = cols
        SIZE[1] = self._rows = rows
        self._data_from_app_queue = asyncio.Queue()
        self._data_to_app_queue = asyncio.Queue()

    def watch__ready(self, value: bool):
        if value:
            SIZE[0] = int(self._cols)
            SIZE[1] = int(self._rows)
            self._tasks = [
                asyncio.create_task(self._run()),
                asyncio.create_task(self._send_data()),
                asyncio.create_task(self._recv_data()),
            ]

    def watch__data_to_app(self, data: str):
        if data:
            self._data_to_app_queue.put_nowait(data)

    async def _send_data(self):
        while True:
            data = await self._data_from_app_queue.get()
            self._data_from_app = ""
            self._data_from_app = data.decode()

    async def _recv_data(self):
        while True:
            data = await self._data_to_app_queue.get()
            if not data:
                continue
            data = data.encode()
            packet_type = b"D"
            packet = b"%s%s%s" % (packet_type, len(data).to_bytes(4, "big"), data)
            self._app._driver._stdin_queue.put_nowait(packet)

    async def _run(self):
        self._tasks.append(asyncio.create_task(self._app.run_async()))
        while True:
            if self._app._driver is not None:
                break
            await asyncio.sleep(0)

        for _ in range(10):
            line = []
            while True:
                data = await self._app._driver._stdout_queue.get()
                line.append(data)
                if data[-1] == 10:  # "\n"
                    break
            line = b"".join(line)
            if not line:
                break
            if line == b"__GANGLION__\n":
                ready = True
                break
        if ready:
            META = 77  # b"M"
            DATA = 68  # b"D"
            while True:
                data = await self._app._driver._stdout_queue.get()
                type_bytes = data[0]
                size_bytes = data[1:5]
                size = int.from_bytes(size_bytes, "big")
                data = data[5:5 + size]
                if type_bytes == DATA:
                    self._data_from_app_queue.put_nowait(data)
                elif type_bytes == META:
                    meta_data = json.loads(data)
                    #if meta_data.get("type") == "exit":
                    #    await self.send_meta({"type": "exit"})
                    #else:
                    #    await on_meta(json.loads(data))



from __future__ import annotations

from ypywidgets import Widget, reactive
from ypywidgets.comm import CommWidget


class SwitchModel(Widget):
    value = reactive(False)

    def __init__(
        self,
        value: bool | None = None,
        ydoc=None,
    ) -> None:
        super().__init__(ydoc)
        if value is not None:
            self.value = value

    def toggle(self):
        self.value = not self.value


class Switch(CommWidget, SwitchModel):

    def __init__(self, value: bool = False) -> None:
        CommWidget.__init__(self)
        SwitchModel.__init__(self, value)



from __future__ import annotations

from ypywidgets import Widget, reactive
from ypywidgets.comm import CommWidget


class PlotextModel(Widget):

    _clear_data = reactive(False)
    _args = reactive([])
    _scatter = reactive(False)
    _plot = reactive(False)
    _title = reactive("")

    def __init__(self, ydoc=None) -> None:
        super().__init__(ydoc)

    def clear_data(self) -> None:
        self._clear_data = False
        self._clear_data = True

    def scatter(self, *args) -> None:
        self._args = list(args)
        self._scatter = False
        self._scatter = True

    def plot(self, *args) -> None:
        self._args = list(args)
        self._plot = False
        self._plot = True

    def title(self, value: str) -> None:
        self._title = value


class Plotext(CommWidget, PlotextModel):

    def __init__(self) -> None:
        CommWidget.__init__(self)
        PlotextModel.__init__(self)



from textual.widgets import Switch as TextualSwitch

from ..switch import SwitchModel


class Switch(TextualSwitch):

    def __init__(self, model: SwitchModel) -> None:
        super().__init__()
        self.ymodel = model
        model.watch_value = self._update_value
        self._update_value(model.value)

    def watch_value(self, value: bool) -> None:
        super().watch_value(value)
        if value != self.ymodel.value:
            self.ymodel.value = value

    def _update_value(self, value: bool) -> None:
        if value != self.value:
            self.value = value



from textual.widgets import Button as TextualButton

from ..button import ButtonModel


class Button(TextualButton):

    def __init__(self, model: ButtonModel) -> None:
        super().__init__()
        self.ymodel = model
        model.watch__press = self.action_press
        model.watch_label = self._set_label
        model.watch_variant = self._set_variant
        model.watch_disabled = self._set_disabled
        self._set_label(model.label)
        self._set_variant(model.variant)
        self._set_disabled(model.disabled)

    def on_button_pressed(self, event) -> None:
        self.ymodel._pressed = not self.ymodel._pressed

    def _set_label(self, value: str) -> None:
        if value != self.label:
            self.label = value

    def _set_variant(self, value: str) -> None:
        if value != self.variant:
            self.variant = value

    def _set_disabled(self, value: str) -> None:
        if value != self.disabled:
            self.disabled = value



import asyncio
from functools import lru_cache
from typing import Callable

import pyte
from rich.color import Color, parse_rgb_hex
from rich.console import RenderableType
from rich.segment import Segment
from rich.style import Style
from textual import events
from textual.widget import Widget as TextualWidget
from wcwidth import wcwidth as _wcwidth

from ..widget import WidgetModel


wcwidth: Callable[[str], int] = lru_cache(maxsize=4096)(_wcwidth)


CTRL_KEYS = {
    "left": "\u001b[D",
    "right": "\u001b[C",
    "up": "\u001b[A",
    "down": "\u001b[B",
}


class PyteDisplay:
    def __init__(self, screen):
        self._screen = screen
        self._cache = {}

    def __rich_console__(self, console, options):
        if self._screen is None:
            return

        dirty_lines = {y: self._render_line(self._screen.buffer[y]) for y in self._screen.dirty}
        self._screen.dirty.clear()
        init = not self._cache
        for y in range(self._screen.lines):
            dirty_line = y in dirty_lines
            if init or dirty_line:
                self._cache[y] = cache = []
                if dirty_line:
                    line = dirty_lines[y]
                else:
                    line = self._render_line(self._screen.buffer[y])
                for char in line:
                    color = None if char.fg == "default" else Color.from_triplet(parse_rgb_hex(char.fg))
                    bgcolor = None if char.bg == "default" else Color.from_triplet(parse_rgb_hex(char.bg))
                    cache.append(
                        Segment(
                            char.data,
                            Style(
                                color=color,
                                bgcolor=bgcolor,
                                bold=char.bold,
                                italic=char.italics,
                                underline=char.underscore,
                                blink=char.blink,
                                strike=char.strikethrough,
                                reverse=char.reverse,
                            )
                        )
                    )
                cache.append(Segment("\n"))

        for y in range(self._screen.lines):
            for segment in self._cache[y]:
                yield segment

    def _render_line(self, line):
        is_wide_char = False
        for x in range(self._screen.columns):
            if is_wide_char:
                is_wide_char = False
                continue
            char = line[x].data
            assert sum(map(wcwidth, char[1:])) == 0
            is_wide_char = wcwidth(char[0]) == 2
            yield line[x]


class Terminal(TextualWidget, can_focus=True):
    DEFAULT_CSS = """
    Terminal {
        height: 1fr;
    }
    """

    def __init__(self, send_queue, recv_queue):
        super().__init__()
        self._send_queue = send_queue
        self._recv_queue = recv_queue
        self._display = PyteDisplay(None)
        self._tasks = [asyncio.create_task(self._recv())]
        self._ready_event = asyncio.Event()

    def render(self) -> RenderableType:
        return self._display

    def on_resize(self, event: events.Resize):
        self.do_resize(event.size.width, event.size.height)

    def do_resize(self, cols: int, rows: int):
        if self.ymodel._cols != cols:
            self.ymodel._cols = cols
        if self.ymodel._rows != rows:
            self.ymodel._rows = rows
        self.ymodel._ready = True
        self._ready_event.set()
        self._screen = pyte.Screen(cols, rows)
        self._stream = pyte.Stream(self._screen)

    async def on_event(self, event: events.Event) -> None:
        await super().on_event(event)
        if isinstance(event, events.Key):
            char = CTRL_KEYS.get(event.key) or event.character
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return
        if isinstance(event, events.MouseMove):
            char = f"\x1b[<35;{event.x};{event.y}M"
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return
        if isinstance(event, events.MouseDown):
            char = f"\x1b[<0;{event.x};{event.y}M"
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return
        if isinstance(event, events.MouseUp):
            char = f"\x1b[<0;{event.x};{event.y}m"
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return

    async def _recv(self):
        while True:
            data = await self._recv_queue.get()
            if not data:
                continue
            self._stream.feed(data)
            self._display = PyteDisplay(self._screen)
            self.refresh()


class Widget(Terminal):

    def __init__(self, model: WidgetModel) -> None:
        self.ymodel = model
        self._data_from_app_queue = asyncio.Queue()
        self._data_to_app_queue = asyncio.Queue()
        super().__init__(self._data_to_app_queue, self._data_from_app_queue)
        model.watch__data_from_app = self._to_terminal
        self._tasks = [asyncio.create_task(self._from_terminal())]

    def _to_terminal(self, data: str) -> None:
        self._data_from_app_queue.put_nowait(data)

    async def _from_terminal(self):
        while True:
            data = await self._data_to_app_queue.get()
            self.ymodel._data_to_app = data



from textual_plotext import PlotextPlot

from ..plotext import PlotextModel


class Plotext(PlotextPlot):

    def __init__(self, model: PlotextModel) -> None:
        super().__init__()
        self.ymodel = model
        model.watch__clear_data = self.clear_data
        model.watch__scatter = self.scatter
        model.watch__plot = self.plot
        model.watch__title = self.title
        self.clear_data()
        self.scatter()
        self.plot()
        self.title(model._title)

    def clear_data(self) -> None:
        if self.ymodel._clear_data:
            self.plt.clear_data()
            self.refresh()

    def title(self, value: str) -> None:
        self.plt.title(value)
        self.refresh()

    def scatter(self) -> None:
        if self.ymodel._scatter:
            self.plt.clear_data()
            self.plt.scatter(*self.ymodel._args)
            self.refresh()

    def plot(self) -> None:
        if self.ymodel._plot:
            self.plt.clear_data()
            self.plt.plot(*self.ymodel._args)
            self.refresh()
